@@ -1,4 +1,8 @@
-//! The goal of this crate is to offer an easy way for bevy games to load all their assets.
+//! The goal of this crate is to offer an easy way for bevy games to load all their assets in a loading State.
+//!
+//! `bevy_asset_loader` introduces the derivable trait [AssetCollection]. Structs with asset handles
+//! can be automatically loaded during a configurable loading [State]. Afterwards they will be inserted as
+//! resources containing loaded handles and the plugin will switch to a second configurable [State].
 //!
 //! ```edition2018
 //! # use bevy_asset_loader::{AssetLoader, AssetCollection};
@@ -13,7 +17,9 @@
 //!     app
 //!         .add_state(GameState::Loading)
 //!         //.add_plugins(DefaultPlugins)
-//!         .add_system_set(SystemSet::on_update(GameState::Next).with_system(use_asset_handles.system()))
+//!         .add_system_set(SystemSet::on_update(GameState::Next)
+//!             .with_system(use_asset_handles.system())
+//!         )
 //!         # .add_plugins(MinimalPlugins)
 //!         # .add_plugin(AssetPlugin::default())
 //!         # .set_runner(|mut app| app.schedule.run(&mut app.world))
@@ -22,10 +28,10 @@
 //!
 //! #[derive(AssetCollection)]
 //! struct AudioAssets {
-//!     #[asset(path = "walking.ogg")]
-//!     walking: Handle<AudioSource>,
-//!     #[asset(path = "flying.ogg")]
-//!     flying: Handle<AudioSource>
+//!     #[asset(path = "audio/background.ogg")]
+//!     background: Handle<AudioSource>,
+//!     #[asset(path = "audio/plop.ogg")]
+//!     plop: Handle<AudioSource>
 //! }
 //!
 //! #[derive(AssetCollection)]
@@ -39,7 +45,7 @@
 //! // since this function runs in [MyState::Next], we know our assets are
 //! // loaded and [MyAudioAssets] is a resource
 //! fn use_asset_handles(audio_assets: Res<AudioAssets>, audio: Res<Audio>) {
-//!     audio.play(audio_assets.flying.clone());
+//!     audio.play(audio_assets.background.clone());
 //! }
 //!
 //! #[derive(Clone, Eq, PartialEq, Debug, Hash)]
@@ -55,11 +61,12 @@
 pub use bevy_asset_loader_derive::AssetCollection;
 
 use bevy::app::AppBuilder;
-use bevy::asset::{AssetServer, HandleId, HandleUntyped, LoadState};
+use bevy::asset::{AssetServer, HandleUntyped, LoadState};
 use bevy::ecs::component::Component;
+use bevy::ecs::prelude::IntoExclusiveSystem;
 use bevy::ecs::schedule::State;
 use bevy::ecs::system::IntoSystem;
-use bevy::prelude::{Commands, Res, ResMut, SystemSet};
+use bevy::prelude::{Commands, FromWorld, Res, ResMut, SystemSet, World};
 use std::fmt::Debug;
 use std::hash::Hash;
 use std::marker::PhantomData;
@@ -87,7 +94,7 @@ pub trait AssetCollection: Component {
 }
 
 struct LoadingAssetHandles<A: Component> {
-    handles: Vec<HandleId>,
+    handles: Vec<HandleUntyped>,
     marker: PhantomData<A>,
 }
 
@@ -97,9 +104,8 @@ struct AssetLoaderConfiguration<T> {
 }
 
 fn start_loading<Assets: AssetCollection>(mut commands: Commands, asset_server: Res<AssetServer>) {
-    let mut handles = Assets::load(&asset_server);
     commands.insert_resource(LoadingAssetHandles {
-        handles: handles.drain(..).map(|handle| handle.id).collect(),
+        handles: Assets::load(&asset_server),
         marker: PhantomData::<Assets>,
     })
 }
@@ -112,7 +118,8 @@ fn check_loading_state<T: Component + Debug + Clone + Eq + Hash, Assets: AssetCo
     loading_asset_handles: Option<Res<LoadingAssetHandles<Assets>>>,
 ) {
     if let Some(loading_asset_handles) = loading_asset_handles {
-        let load_state = asset_server.get_group_load_state(loading_asset_handles.handles.clone());
+        let load_state = asset_server
+            .get_group_load_state(loading_asset_handles.handles.iter().map(|handle| handle.id));
         if load_state == LoadState::Loaded {
             commands.insert_resource(Assets::create(&asset_server));
             commands.remove_resource::<LoadingAssetHandles<Assets>>();
@@ -128,6 +135,11 @@ fn check_loading_state<T: Component + Debug + Clone + Eq + Hash, Assets: AssetCo
     }
 }
 
+fn post_process<Asset: FromWorld + Component>(world: &mut World) {
+    let asset = Asset::from_world(world);
+    world.insert_resource(asset);
+}
+
 /// A Bevy plugin to configure automatic asset loading
 ///
 /// ```edition2018
@@ -152,7 +164,7 @@ fn check_loading_state<T: Component + Debug + Clone + Eq + Hash, Assets: AssetCo
 /// }
 ///
 /// fn play_audio(audio_assets: Res<AudioAssets>, audio: Res<Audio>) {
-///     audio.play(audio_assets.flying.clone());
+///     audio.play(audio_assets.background.clone());
 /// }
 ///
 /// #[derive(Clone, Eq, PartialEq, Debug, Hash)]
@@ -163,8 +175,8 @@ fn check_loading_state<T: Component + Debug + Clone + Eq + Hash, Assets: AssetCo
 ///
 /// #[derive(AssetCollection)]
 /// pub struct AudioAssets {
-///     #[asset(path = "audio/flying.ogg")]
-///     pub flying: Handle<AudioSource>,
+///     #[asset(path = "audio/background.ogg")]
+///     pub background: Handle<AudioSource>,
 /// }
 ///
 /// #[derive(AssetCollection)]
@@ -179,6 +191,7 @@ pub struct AssetLoader<T> {
     next: T,
     load: SystemSet,
     check: SystemSet,
+    post_process: SystemSet,
     collection_count: usize,
 }
 
@@ -214,8 +227,8 @@ where
     /// # }
     /// # #[derive(AssetCollection)]
     /// # pub struct AudioAssets {
-    /// #     #[asset(path = "audio/flying.ogg")]
-    /// #     pub flying: Handle<AudioSource>,
+    /// #     #[asset(path = "audio/background.ogg")]
+    /// #     pub background: Handle<AudioSource>,
     /// # }
     /// # #[derive(AssetCollection)]
     /// # pub struct TextureAssets {
@@ -229,7 +242,8 @@ where
         Self {
             next,
             load: SystemSet::on_enter(load.clone()),
-            check: SystemSet::on_update(load),
+            check: SystemSet::on_update(load.clone()),
+            post_process: SystemSet::on_exit(load),
             collection_count: 0,
         }
     }
@@ -261,8 +275,8 @@ where
     /// # }
     /// # #[derive(AssetCollection)]
     /// # pub struct AudioAssets {
-    /// #     #[asset(path = "audio/flying.ogg")]
-    /// #     pub flying: Handle<AudioSource>,
+    /// #     #[asset(path = "audio/background.ogg")]
+    /// #     pub background: Handle<AudioSource>,
     /// # }
     /// # #[derive(AssetCollection)]
     /// # pub struct TextureAssets {
@@ -280,6 +294,56 @@ where
         self
     }
 
+    /// Add any [FromWorld] resource to be inititlized after all asset collections are loaded.
+    /// ```edition2018
+    /// # use bevy_asset_loader::{AssetLoader, AssetCollection};
+    /// # use bevy::prelude::*;
+    /// # use bevy::asset::AssetPlugin;
+    /// # fn main() {
+    ///     let mut app = App::build();
+    ///     AssetLoader::new(GameState::Loading, GameState::Menu)
+    ///         .with_collection::<TextureForAtlas>()
+    ///         .post_process::<TextureAtlasFromWorld>()
+    ///         .build(&mut app);
+    /// #   app
+    /// #       .add_state(GameState::Loading)
+    /// #       .add_plugins(MinimalPlugins)
+    /// #       .add_plugin(AssetPlugin::default())
+    /// #       .set_runner(|mut app| app.schedule.run(&mut app.world))
+    /// #       .run();
+    /// # }
+    /// # #[derive(Clone, Eq, PartialEq, Debug, Hash)]
+    /// # enum GameState {
+    /// #     Loading,
+    /// #     Menu
+    /// # }
+    /// # struct TextureAtlasFromWorld {
+    /// #     atlas: Handle<TextureAtlas>
+    /// # }
+    /// # impl FromWorld for TextureAtlasFromWorld {
+    /// #     fn from_world(world: &mut World) -> Self {
+    /// #         let cell = world.cell();
+    /// #         let assets = cell.get_resource::<TextureForAtlas>().expect("TextureForAtlas not loaded");
+    /// #         let mut atlases = cell.get_resource_mut::<Assets<TextureAtlas>>().expect("TextureAtlases missing");
+    /// #         TextureAtlasFromWorld {
+    /// #             atlas: atlases.add(TextureAtlas::from_grid(assets.array.clone(), Vec2::new(250., 250.), 1, 4))
+    /// #         }
+    /// #     }
+    /// # }
+    /// # #[derive(AssetCollection)]
+    /// # pub struct TextureForAtlas {
+    /// #     #[asset(path = "textures/female_adventurer.ogg")]
+    /// #     pub array: Handle<Texture>,
+    /// # }
+    /// ```
+    pub fn post_process<A: FromWorld + Component>(mut self) -> Self {
+        self.post_process = self
+            .post_process
+            .with_system(post_process::<A>.exclusive_system());
+
+        self
+    }
+
     /// Finish configuring the [AssetLoader]
     ///
     /// Calling this function is required to set up the asset loading.
@@ -307,8 +371,8 @@ where
     /// # }
     /// # #[derive(AssetCollection)]
     /// # pub struct AudioAssets {
-    /// #     #[asset(path = "audio/flying.ogg")]
-    /// #     pub flying: Handle<AudioSource>,
+    /// #     #[asset(path = "audio/background.ogg")]
+    /// #     pub background: Handle<AudioSource>,
     /// # }
     /// # #[derive(AssetCollection)]
     /// # pub struct TextureAssets {
@@ -321,6 +385,7 @@ where
     pub fn build(self, app: &mut AppBuilder) {
         app.add_system_set(self.load)
             .add_system_set(self.check)
+            .add_system_set(self.post_process)
             .insert_resource(AssetLoaderConfiguration::<T> {
                 count: self.collection_count,
                 next: self.next,
